local WINDOW_HEIGHT = 16 * 13
local WINDOW_WIDTH = 16 * 16
local GAME_END_TIME = 2.7
local MIN_ITEM_SPAWN = 5
local MAX_ITEM_SPAWN = 9
local ITEM_SPAWN_ITEM_QUEUE = 3
local ITEM_SPAWN_POS_QUEUE = 7

local TEXT_BANK = {
	["selMap"] = "Welcome to Top Battle,\nPlease select a map.",
	["rebind"] = "Keyboard rebind.",
	["bindin"] = "Please press a key.",
}

function love.load()
	-- set the default filter
	love.graphics.setDefaultFilter("nearest", "nearest")
	-- setup the save directory
	love.filesystem.createDirectory("sound/music_title")
	love.filesystem.createDirectory("sound/music_game")
	-- get graphics limits
	local grapLim = love.graphics.getSystemLimits()
	-- multiCanvasLim = grapLim["multicanvas"] -- ugh single pass rendering is buggy (get both players inventory full, let one player drop a item, move the other player onto the item)
	multiCanvasLim = 1
	-- load modules
	card_module = require("modules/graphics/card")
	mapCanvas_module = require("modules/graphics/mapCanvas")
	keyboard_module = require("modules/input/keyboard")
	text_module = require("modules/graphics/text")
	playerManager_module = require("modules/game/playerManager")
	mapLoader_module = require("modules/map/mapLoader")
	entityManager_module = require("modules/entity/entityManager")
	itemManager_module = require("modules/game/itemManager")
	inventoryManager_module = require("modules/game/inventoryManager")
	spritePool_module = require("modules/graphics/spritePool")
	fileSet_module = require("modules/file/fileSet")
	shaderManager_module = require("modules/graphics/shaderManager")
	meleeManager_module = require("modules/game/meleeManager")
	projectileManager_module = require("modules/game/projectileManager")
	timerManager_module = require("modules/game/timerManager")
	sound_module = require("modules/sound/sound")
	mapTile_module = require("modules/map/mapTile")
	settingFile_module = require("modules/file/settingFile")
	-- define variables
    gameState = 2
	--[[
		game states:
		0 = main menu 1 = rebind 2 = prepare for main menu 3 = main game 4 = prepare for main game
		5 = reset key bindings
	--]]
	mapSelect = 0
	maps = love.filesystem.getDirectoryItems("maps")
	pause = false -- also a rebind state
	-- scale = nil -- setting
	winningPlayer = 0
	endTime = 0
	mapTimerID = 0 -- also a select index for the keyboard rebind mode
	placedItems = {} -- 2
	pickedPositions = {} -- 7
	spawnTimerID = nil
	bindResetTimerID = nil
	musicChannelID = nil
	-- playerIDs
	player1ID = nil
	player2ID = nil
	-- define functions
	loadMap = function(mapName)
		mapLoader_module.loadMap(mapName)
		mapTile_module.refreshTiles()
		mapCanvas_module.renderMap()
	end
	-- load setting defaults
	settingFile_module.setDefaults({
		["windowSize"] = 4,
		["keyPlr1Up"] = "w",
		["keyPlr1Down"] = "s",
		["keyPlr1Left"] = "a",
		["keyPlr1Right"] = "d",
		["keyPlr1Fire"] = "j",
		["keyPlr1Inv"] = "k",
		["keyPlr1Aim"] = "l",
		["keyPlr2Up"] = "up",
		["keyPlr2Down"] = "down",
		["keyPlr2Left"] = "left",
		["keyPlr2Right"] = "right",
		["keyPlr2Fire"] = "kp4",
		["keyPlr2Inv"] = "kp5",
		["keyPlr2Aim"] = "kp6",
		["keyPause"] = "escape",
		["keyScrSize"] = "f1",
	})
	settingFile_module.setToDefault()
	-- load the setting file
	settingFile_module.load()
	-- set variables to settings
	scale = settingFile_module.getSetting("windowSize")
	-- set window size
	love.window.setMode(WINDOW_WIDTH * scale, WINDOW_HEIGHT * scale)
	-- setup keys
	keyboard_module.setKey(1, settingFile_module.getSetting("keyPlr1Up")) -- player 1 up
	keyboard_module.setKey(2, settingFile_module.getSetting("keyPlr1Left")) -- player 1 left
	keyboard_module.setKey(3, settingFile_module.getSetting("keyPlr1Down")) -- player 1 down
	keyboard_module.setKey(4, settingFile_module.getSetting("keyPlr1Right")) -- player 1 right
	keyboard_module.setKey(5, settingFile_module.getSetting("keyPlr1Fire")) -- player 1 attack
	keyboard_module.setKey(6, settingFile_module.getSetting("keyPlr1Aim")) -- player 1 aim
	keyboard_module.setKey(7, settingFile_module.getSetting("keyPlr1Inv")) -- player 1 inventory
	keyboard_module.setKey(8, settingFile_module.getSetting("keyPlr2Up")) -- player 2 up
	keyboard_module.setKey(9, settingFile_module.getSetting("keyPlr2Left")) -- player 2 left
	keyboard_module.setKey(10, settingFile_module.getSetting("keyPlr2Down")) -- player 2 down
	keyboard_module.setKey(11, settingFile_module.getSetting("keyPlr2Right")) -- player 2 right
	keyboard_module.setKey(12, settingFile_module.getSetting("keyPlr2Fire")) -- player 2 attack
	keyboard_module.setKey(13, settingFile_module.getSetting("keyPlr2Aim")) -- player 2 aim
	keyboard_module.setKey(14, settingFile_module.getSetting("keyPlr2Inv")) -- player 2 inventory
	keyboard_module.setKey(15, settingFile_module.getSetting("keyPause")) -- pause
	keyboard_module.setKey(16, settingFile_module.getSetting("keyScrSize")) -- change screen size
	-- reset key binding keys
	keyboard_module.setKey(100, "y")
	keyboard_module.setKey(101, "n")
	-- load the pause graphic
	pauseGraphic = love.graphics.newImage("res/pauseGraphic.png")
	-- filter the map
	for i = #maps, 1, -1 do
		local mapName = maps[i]
		if not love.filesystem.getInfo("maps/"..mapName)[1] == "directory" or
		mapName == "title" or 
		mapName == "keyRebind" then
			table.remove(maps, i)
		end
	end
	maps[0] = "title"
	-- set the randomizer
	math.randomseed(os.time())
	math.random();math.random();math.random()
	-- load audio files
	sound_module.load("res/sfx/pickup.wav", true, "pickup")
	sound_module.load("res/sfx/hit.wav", true, "hit")
	sound_module.load("res/sfx/shoot.wav", true, "shoot")
	sound_module.load("res/sfx/meleeShoot.wav", true, "meleeShoot")
	-- load shaders
	shaderManager_module.loadShaders({
		["entityShader"] = {
			multiCanvasLim,
			{
				[1] = "multiPass",
				["default"] = "singlePass"
			}
		}
	})
	-- load templates and item types
	fileSet_module.loadTemplates()
	itemManager_module.loadItemTypes()
    -- load file sets and item sprites
	fileSet_module.loadSets("res/fileSet.txt")
	-- load graphics
	cursorGraphic = spritePool_module.load("res/inventory/cursor.png")
	equipGraphic = spritePool_module.load("res/inventory/equipMark.png")
	-- prepare the text module
	text_module.prepare()
	-- render the inventory interface
	mapCanvas_module.renderInventoryInterface()
	mapCanvas_module.renderInventoryOverlay()
	-- set the text
	mapCanvas_module.setText(TEXT_BANK["selMap"])
	-- rebind canvas so the text won't look ugly
	textCanvas = love.graphics.newCanvas()
end

function love.quit()
	print("Saving settings")
	settingFile_module.save()
	print("Done, now closing")
end

local keyd = { -- key rebind setting keys
	"keyPause", "keyScrSize",
	"keyPlr1Up", "keyPlr2Up",
	"keyPlr1Down", "keyPlr2Down",
	"keyPlr1Left", "keyPlr2Left",
	"keyPlr1Right", "keyPlr2Right",
	"keyPlr1Fire", "keyPlr2Fire",
	"keyPlr1Inv", "keyPlr2Inv",
	"keyPlr1Aim", "keyPlr2Aim",
	15, 16,
	1, 8,
	3, 10,
	2, 9,
	4, 11,
	5, 12,
	7, 14,
	6, 13,
}
function love.keypressed(key)
	if gameState == 1 and pause == true then
		settingFile_module.setSetting(keyd[mapTimerID + 1], key)
		keyboard_module.setKey(keyd[mapTimerID + 17], key)
		mapCanvas_module.setText(TEXT_BANK["rebind"])
		pause = nil
	end
end

function love.update(dt)
	keyboard_module.update()
	timerManager_module.update(dt)
	sound_module.update()
	if keyboard_module.isPressed(16) and not (gameState == 1 and pause == nil) then
		scale = 5 - ((5 - scale) % 4 + 1)
		love.window.setMode(WINDOW_WIDTH * scale, WINDOW_HEIGHT * scale)
		mapCanvas_module.renderMap()
		mapCanvas_module.forceTextRedraw()
		settingFile_module.setSetting("windowSize", scale)
	end
	if gameState == 0 then -- main menu
		if keyboard_module.isPressed(2) then -- prev
			if mapSelect > 0 then
				mapSelect = mapSelect - 1
				loadMap(maps[mapSelect])
				if mapSelect > 0 then
					mapCanvas_module.setText(maps[mapSelect])
				else
					mapCanvas_module.setText(TEXT_BANK["selMap"])
				end
			end
		elseif keyboard_module.isPressed(4) then -- next
			if mapSelect < #maps then
				mapSelect = mapSelect + 1
				loadMap(maps[mapSelect])
				mapCanvas_module.setText(maps[mapSelect])
			end
		elseif keyboard_module.isPressed(5) then -- start
			if mapSelect ~= 0 then
				sound_module.folderPlay("music", musicChannelID, "res/music_game", "sound/music_game")
				gameState = 4
			end
		elseif keyboard_module.isPressed(15) then -- go to rebind
			loadMap("keyRebind")
			mapCanvas_module.setText(TEXT_BANK["rebind"])
			gameState = 1
		end
		if love.keyboard.isDown("lctrl") and love.keyboard.isDown("tab") and love.keyboard.isDown("delete") then
			if bindResetTimerID then
				if timerManager_module.getState(bindResetTimerID) == 2 then
					loadMap("keyRebind")
					gameState = 5
					timerManager_module.destroy(bindResetTimerID)
					bindResetTimerID = nil
				end
			else
				bindResetTimerID = timerManager_module.new(3, false)
			end
		else
			if bindResetTimerID then
				timerManager_module.destroy(bindResetTimerID)
				bindResetTimerID = nil
			end
		end
	elseif gameState == 1 then -- rebind
		if pause == false then
			if keyboard_module.isPressed(1) then -- up
				mapTimerID = mapTimerID - 2
				if mapTimerID < 0 then
					mapTimerID = mapTimerID + 16
				end
			elseif keyboard_module.isPressed(2) then -- left
				mapTimerID = mapTimerID - 1
				if mapTimerID < 0 then
					mapTimerID = mapTimerID + 16
				end
			elseif keyboard_module.isPressed(3) then -- down
				mapTimerID = mapTimerID + 2
				if mapTimerID > 15 then
					mapTimerID = mapTimerID - 16
				end
			elseif keyboard_module.isPressed(4) then -- right
				mapTimerID = mapTimerID + 1
				if mapTimerID > 15 then
					mapTimerID = mapTimerID - 16
				end
			elseif keyboard_module.isPressed(5) then -- start rebind
				mapCanvas_module.setText(TEXT_BANK["bindin"])
				pause = true
			elseif keyboard_module.isPressed(15) then -- go to menu
				loadMap(maps[mapSelect])
				if mapSelect > 0 then
					mapCanvas_module.setText(maps[mapSelect])
				else
					mapCanvas_module.setText(TEXT_BANK["selMap"])
				end
				mapTimerID = 0
				pause = false
				gameState = 0
			end
		elseif pause == nil then
			pause = false
		end
	elseif gameState == 2 then -- prepare for main menu
		loadMap("title")
		mapCanvas_module.setText(TEXT_BANK["selMap"])
		if not musicChannelID then
			musicChannelID = sound_module.new()
		end
		sound_module.folderPlay("music", musicChannelID, "res/music_title", "sound/music_title")
		gameState = 0
	elseif gameState == 3 then -- main game
		if winningPlayer ~= 0 then
			if winningPlayer > 0 then
				card_module.selectCard(winningPlayer)
				card_module.start()
				timerManager_module.pause(mapTimerID)
				winningPlayer = -1
			else
				card_module.update()
			end
			if card_module.hasFinished() then
				if endTime >= GAME_END_TIME then
					sound_module.stopAll()
					sound_module.clear()
					card_module.reset()
					timerManager_module.clear()
					meleeManager_module.clear()
					projectileManager_module.clear()
					playerManager_module.clear()
					inventoryManager_module.clear()
					itemManager_module.clear()
					player1ID = nil
					player2ID = nil
					entityManager_module.clear()
					playerManager_module.finish()
					mapSelect = 0
					gameState = 2
					winningPlayer = 0
					endTime = 0
					mapTimerID = 0
					musicChannelID = nil
					-- placedItems = {}
					-- pickedPositions = {}
				else
					endTime = endTime + dt
				end
			end
			return
		end
		if keyboard_module.isPressed(15) then
			pause = not pause
			if pause then
				timerManager_module.pause(mapTimerID)
				sound_module.pauseAll()
			else
				timerManager_module.resume(mapTimerID)
				sound_module.playAll()
			end
		end
		if not pause then
			playerManager_module.update(dt)
			itemManager_module.update()
			inventoryManager_module.update()
			meleeManager_module.update()
			projectileManager_module.update()
			entityManager_module.update()
			-- win code
			local playerWinCond = 
				(playerManager_module.getHealth(player1ID) <= 0 and 2 or 0) +
				(playerManager_module.getHealth(player2ID) <= 0 and 1 or 0)
			if playerWinCond ~= 0 then
				winningPlayer = playerWinCond
			elseif timerManager_module.getState(mapTimerID) == 2 then
				winningPlayer = 4
			end
			-- item spawn code
			if timerManager_module.getState(spawnTimerID) == 2 then
				local x, y = 0, 0
				local stop = true
				while true do
					x, y = math.random(0, 15), math.random(0, 10)
					-- check if the position selected isn't already picked
					for i = 1, #pickedPositions, 2 do
						if pickedPositions[i] == x and pickedPositions[i + 1] == y then
							stop = false
						end
					end
					-- spawn if it's not and if the fore tile in that position is blank
					if stop and mapLoader_module.getForeTile(x + 1, y + 1) == "00" then
						while true do
							local selectedItem = fileSet_module.getRandomName("item")
							-- check if the item selected isn't already picked
							for i = 1, #placedItems do
								if placedItems[i] == selectedItem then
									stop = false
									break
								end
							end
							-- spawn if it's not
							if stop then
								itemManager_module.new(selectedItem, 16 * x, 16 * y)
								-- push into the item and position queues
								if #placedItems >= ITEM_SPAWN_ITEM_QUEUE then
									table.remove(placedItems, 1)
								end
								placedItems[#placedItems + 1] = selectedItem
								if #pickedPositions >= ITEM_SPAWN_POS_QUEUE * 2 then
									table.remove(pickedPositions, 1)
									table.remove(pickedPositions, 1)
								end
								pickedPositions[#pickedPositions + 1] = x
								pickedPositions[#pickedPositions + 1] = y
								break
							end
							stop = true
						end
						break
					end
					stop = true
				end
				timerManager_module.setTime(spawnTimerID, math.random(MIN_ITEM_SPAWN, MAX_ITEM_SPAWN))
				timerManager_module.resume(spawnTimerID)
			end
		end
	elseif gameState == 4 then -- prepare for state 3
		-- spawn players
		playerManager_module.prepare()
		player1ID = playerManager_module.new(mapLoader_module.getPlayerPos(1), {0, 0, 1}, {1, 3, 2, 4, 7, 5, 6})
		player2ID = playerManager_module.new(mapLoader_module.getPlayerPos(2), {1, 0, 0}, {8, 10, 9, 11, 14, 12, 13})
		-- set timer
		local mtm, mts = mapLoader_module.getMapTime()
		mapTimerID = timerManager_module.new(60 * mtm + mts)
		-- set item spawn timer
		spawnTimerID = timerManager_module.new(math.random(MIN_ITEM_SPAWN, MAX_ITEM_SPAWN))
		-- set game state
		gameState = 3
	elseif gameState == 5 then -- reset key bindings
		if keyboard_module.isPressed(100) then -- yes
			for i = 1, #keyd / 2 do
				settingFile_module.setToDefault(keyd[i])
				keyboard_module.setKey(keyd[i + 16], settingFile_module.getSetting(keyd[i]))
			end
			loadMap(maps[mapSelect])
			gameState = 0
		elseif keyboard_module.isPressed(101) then -- no
			loadMap(maps[mapSelect])
			gameState = 0
		end
	end
end

function love.draw()
	-- scale
	love.graphics.scale(scale, scale)
	-- map
	local map = mapCanvas_module.getMapGraphic()
	love.graphics.draw(map, 0, 0)
	-- interface
	local invenBar = mapCanvas_module.getInventoryInterface()
	love.graphics.draw(invenBar, 0, 176)
	if gameState == 3 then -- game (new system)
		-- inventory overlay
		love.graphics.setBlendMode("alpha", "premultiplied")
		local invenOverlay = mapCanvas_module.getInventoryOverlay()
		love.graphics.draw(invenOverlay, 0, 192)
		love.graphics.setBlendMode("alpha", "alphamultiply")
		-- items
		local player1Inventory = inventoryManager_module.getEntry(player1ID)
		local player2Inventory = inventoryManager_module.getEntry(player2ID)
		for i = 1, 8 do
			local player1Item = itemManager_module.getEntry(player1Inventory[i])
			local player2Item = itemManager_module.getEntry(player2Inventory[i])
			if player1Item then
				love.graphics.draw(spritePool_module.get(fileSet_module.get("item", player1Item[2])[1]), 16 * (i - 1), 192)
			end
			if player2Item then
				love.graphics.draw(spritePool_module.get(fileSet_module.get("item", player2Item[2])[1]), 16 * (i + 7), 192)
			end
		end
		-- inventory cursor
		shaderManager_module.setShader("tint")
		shaderManager_module.sendColor("tint", {109 / 255, 178 / 255, 1})
		love.graphics.draw(spritePool_module.get(cursorGraphic), 16 * (playerManager_module.getInventoryIndex(player1ID) - 1), 192)
		shaderManager_module.sendColor("tint", {1, 167 / 255, 85 / 255})
		love.graphics.draw(spritePool_module.get(cursorGraphic), 16 * (playerManager_module.getInventoryIndex(player2ID) + 7), 192)
		-- equip marks
		local plr1EquSlot = playerManager_module.getWeaponEquip(player1ID)
		local plr2EquSlot = playerManager_module.getWeaponEquip(player2ID)
		local plr1ArmSlot = playerManager_module.getArmourEquip(player1ID)
		local plr2ArmSlot = playerManager_module.getArmourEquip(player2ID)
		-- weapon
		shaderManager_module.sendColor("tint", {132 / 255, 1, 205 / 255})
		love.graphics.setScissor(0, 16 * 12 * scale, 16 * 8 * scale, 16 * scale)
		love.graphics.draw(spritePool_module.get(equipGraphic), (plr1EquSlot - 1) * 16, 192)
		love.graphics.setScissor(16 * 8 * scale, 16 * 12 * scale, 16 * 8 * scale, 16 * scale)
		love.graphics.draw(spritePool_module.get(equipGraphic), (plr2EquSlot + 7) * 16, 192)
		-- armour
		shaderManager_module.sendColor("tint", {186 / 255, 140 / 255, 225 / 255})
		love.graphics.setScissor(0, 16 * 12 * scale, 16 * 8 * scale, 16 * scale)
		love.graphics.draw(spritePool_module.get(equipGraphic), (plr1ArmSlot - 1) * 16, 192)
		love.graphics.setScissor(16 * 8 * scale, 16 * 12 * scale, 16 * 8 * scale, 16 * scale)
		love.graphics.draw(spritePool_module.get(equipGraphic), (plr2ArmSlot + 7) * 16, 192)
		love.graphics.setScissor()
		shaderManager_module.setShader()
		-- health and ammo
		local player1Ammo = 0
		local weaponPointer = playerManager_module.getItemIDFromWeaponEquip(player1ID)
		if itemManager_module.doesExist(weaponPointer) then
			player1Ammo = itemManager_module.getVariable(weaponPointer, "ammo") or 0
		end
		local player2Ammo = 0
		weaponPointer = playerManager_module.getItemIDFromWeaponEquip(player2ID)
		if itemManager_module.doesExist(weaponPointer) then
			player2Ammo = itemManager_module.getVariable(weaponPointer, "ammo") or 0
		end
		text_module.smallNumber(playerManager_module.getHealth(player1ID),    8,   177) -- player 1
		text_module.smallNumber(player1Ammo,                                  8,   185)
		text_module.smallNumberRev(playerManager_module.getHealth(player2ID), 242, 177) -- player 2
		text_module.smallNumberRev(player2Ammo,                               242, 185)
		-- symbols
		text_module.symbol(1, 0, 176)
		text_module.symbol(2, 0, 184)
		text_module.symbol(1, 248, 176)
		text_module.symbol(2, 248, 184)
		-- timer
		local mapTime = timerManager_module.getTimeRounded(mapTimerID)
		text_module.bigNumber(string.format("%02d", math.floor(mapTime / 60)), 114, 177, {48 / 255, 158 / 255, 1})
		text_module.bigNumber(string.format("%02d", mapTime % 60), 129, 177, {1, 158 / 255, 48 / 255})
		-- entities
		entityManager_module.render()
		-- pause graphic
		if pause then
			love.graphics.draw(pauseGraphic, 112, 80)
		end
		-- end card
		card_module.render()
	elseif gameState == 0 then -- main menu
		mapCanvas_module.drawText(0, 16 * 11)
	elseif gameState == 1 then -- rebind
		love.graphics.push("all")
		love.graphics.reset()
		love.graphics.setColor(0, 0, 0)
		love.graphics.setCanvas(textCanvas)
		love.graphics.clear()
		-- text
		love.graphics.printf("Pause", 32, 0, 16 * 8, "center") -- pause
		love.graphics.printf("Change screen size", 16 * 8, 0, 16 * 8, "center") -- screen size
		love.graphics.printf("Player 1", 32, 32, 16 * 8, "center") -- player 1
		love.graphics.printf("Player 2", 16 * 8, 32, 16 * 8, "center") -- player 2
		love.graphics.print("Up:", 0, 16 * 3) -- up
		love.graphics.print("Down:", 0, 16 * 4) -- down
		love.graphics.print("Left:", 0, 16 * 5) -- left
		love.graphics.print("Right:", 0, 16 * 6) -- right
		love.graphics.print("Fire:", 0, 16 * 7) -- fire
		love.graphics.print("Inventory:", 0, 16 * 8) -- inventory
		love.graphics.print("Aim:", 0, 16 * 9) -- aim
		-- boxes
		local index = 0
		local cR, cG, cB = 0.15, 0.15, 0.15
		if pause then -- rebind active
			cG = 0.87
		else -- rebind not active
			cR = 0.87
		end
		for i = 1.9, 10, 1.1 do
			if math.floor(mapTimerID / 2) == index then
				if mapTimerID % 2 == 1 then -- right box
					love.graphics.setColor(0.87, 0.87, 0.87)
					love.graphics.rectangle("fill", 16 * 4, 16 * math.floor(i), 16 * 4, 16) -- left box
					love.graphics.setColor(cR, cG, cB)
					love.graphics.rectangle("fill", 16 * 10, 16 * math.floor(i), 16 * 4, 16) -- right box
				else -- left box
					love.graphics.setColor(cR, cG, cB)
					love.graphics.rectangle("fill", 16 * 4, 16 * math.floor(i), 16 * 4, 16) -- left box
					love.graphics.setColor(0.87, 0.87, 0.87)
					love.graphics.rectangle("fill", 16 * 10, 16 * math.floor(i), 16 * 4, 16) -- right box
				end
			else
				love.graphics.setColor(0.87, 0.87, 0.87)
				love.graphics.rectangle("fill", 16 * 4, 16 * math.floor(i), 16 * 4, 16) -- left box
				love.graphics.rectangle("fill", 16 * 10, 16 * math.floor(i), 16 * 4, 16) -- right box
			end
			love.graphics.setColor(0.23, 0.23, 0.23)
			-- left text
			love.graphics.setScissor(16 * 4, 16 * math.floor(i), 16 * 4, 16)
			love.graphics.printf(settingFile_module.getSetting(keyd[2 * index + 1]), 16 * 4, 16 * math.floor(i), 16 * 4, "center")
			-- right text
			love.graphics.setScissor(16 * 10, 16 * math.floor(i), 16 * 4, 16)
			love.graphics.printf(settingFile_module.getSetting(keyd[2 * index + 2]), 16 * 10, 16 * math.floor(i), 16 * 4, "center")
			love.graphics.setScissor() -- reset scissor
			index = index + 1
		end
		love.graphics.pop()
		love.graphics.draw(textCanvas)
		mapCanvas_module.drawText(0, 16 * 11)
	elseif gameState == 5 then -- reset key bindings
		love.graphics.push("all")
		love.graphics.reset()
		love.graphics.setColor(0, 0, 0)
		love.graphics.setCanvas(textCanvas)
		love.graphics.clear()
		-- text
		love.graphics.printf("Are you sure you want to reset key bindings?", 0, 16 * 4, 16 * 16, "center")
		love.graphics.printf("(Y/N)", 0, 16 * 6, 16 * 16, "center")
		love.graphics.pop()
		love.graphics.draw(textCanvas)
	end
	-- reset coordinate system (automatically resets after the draw function)
	--love.graphics.origin()
	-- pref
	-- local ps = love.graphics.getStats()
	-- for i, v in pairs(ps) do
		-- print(i.." "..v)
	-- end
end
