local entityManager_module = {}

local spritePool_module = require("modules/graphics/spritePool")
local mapLoader_module = require("modules/map/mapLoader")
local shaderManager_module = require("modules/graphics/shaderManager")

local u_error = require("modules/utils/error")

--[[
	entity container:
	{{x, y}, reserved, spriteID, quads, quadSelect, tintR, tintG, tintB, forceX, forceY, {bounding boxes}, collidable,
	anchor tag, anchorX, anchorY, flipX, flipY, {flip boxes}, render layer}
--]]

local entities = {}
local anchorTable = {}

local mapWidth = 16 * 16
local mapHeight = 16 * 11

local screenWidth, screenHeight = love.graphics.getDimensions()
local renderCanvas = love.graphics.newCanvas(screenWidth, screenHeight)
local stencilCanvas= love.graphics.newCanvas(screenWidth, screenHeight, {["format"] = "r8"})

local function _doesExist(entID, m)
	if not entities[entID] then
		u_error(m, "entity object does not exist!")
	end
end

local function py(a,b)return math.abs(a[1] ^ 2 + a[2] ^ 2) < math.abs(b[1] ^ 2 + b[2] ^ 2) end
local edgeDiff = {{0, 0}, {0, 0}, {0, 0}, {0, 0}}
local function puul(ent1X, ent1Y, ent1BB, ent2X, ent2Y, ent2BB, collFunc)
	for _, bb1 in pairs(ent1BB) do
		local aw = bb1[4] - bb1[3]
		local ah = bb1[2] - bb1[1]
		local ax = ent1X + bb1[3]
		local ay = ent1Y + bb1[1]
		for _, bb2 in pairs(ent2BB) do
			local bw = bb2[4] - bb2[3]
			local bh = bb2[2] - bb2[1]
			local bx = ent2X + bb2[3]
			local by = ent2Y + bb2[1]
			if (ax < bx + bw and ax + aw > bx) and (ay < by + bh and ay + ah > by) then
				aw, ah = ax + aw - bx, ay + ah - by
				bw, bh = ax - (bx + bw), ay - (by + bh)
				edgeDiff[1][1], edgeDiff[2][1] = bw, bw
				edgeDiff[3][1], edgeDiff[4][1] = aw, aw
				edgeDiff[1][2], edgeDiff[3][2] = bh, bh
				edgeDiff[2][2], edgeDiff[4][2] = ah, ah
				table.sort(edgeDiff, py)
				local status = collFunc(edgeDiff[1])
				if status then
					return
				end
			end
		end
	end
end

local function boxFlip(up, down, left, right, flipX, flipY)
	local newUp = up
	local newDown = down
	local newLeft = left
	local newRight = right
	if flipX then
		newLeft = 16 - right
		newRight = 16 - left
	end
	if flipY then
		newUp = 16 - down
		newDown = 16 - up
	end
	return newUp, newDown, newLeft, newRight
end

local tileBoundingBox = {{0, 16, 0, 16}}
local function iterateTiles(entPos, entBoundBox, func, foreMap)
	local entTileX = math.floor(entPos[1] / 16) - 1
	local entTileY = math.floor(entPos[2] / 16) - 1
	for tileY = entTileY - 1, entTileY + 2 do
		for tileX = entTileX - 1, entTileX + 2 do
			local tile = foreMap[16 * tileY + tileX + 1]
			if tile and tile ~= "00" then
				puul(entPos[1], entPos[2], entBoundBox, 16 * tileX, 16 * tileY, tileBoundingBox, func)
			end
		end
	end
end

local function scanPush(axis, entPos, entBoundBox, foreMap)
	iterateTiles(entPos, entBoundBox, function(diff)
		entPos[axis] = entPos[axis] - diff[axis]
	end, foreMap)
end

function entityManager_module.new(posX, posY, spriteID)
	local ent = {}
	ent[1] = {posX, posY} -- position
	-- RESERVED
	ent[3] = spriteID or 0 -- sprite ID
	ent[4] = nil -- quads
	ent[5] = 1 -- quad select
	ent[6] = 0 -- tint red
	ent[7] = 0 -- tint green
	ent[8] = 0 -- tint blue
	ent[9] = 0 -- force X
	ent[10]= 0 -- force Y
	ent[11]= {} -- bounding boxes
	ent[12]= true -- collidable
	ent[13]= nil -- anchor tag
	ent[14]= 0 -- anchor offset X
	ent[15]= 0 -- anchor offset Y
	ent[16]= false -- flip X
	ent[17]= false -- flip Y
	ent[18]= {} -- flip boxes
	ent[19]= 1 -- render layer
	local newID = #entities + 1
	entities[newID] = ent
	return newID
end

function entityManager_module.destroy(entID)
	_doesExist(entID, "destroying an entity object")
	entityManager_module.unanchor(entID)
	entities[entID] = nil
end

function entityManager_module.clear()
	for i, _ in pairs(entities) do
		entityManager_module.destroy(i)
	end
end

function entityManager_module.getArray()
	return entities
end

function entityManager_module.getEntry(id)
	return entities[id]
end

function entityManager_module.getCollidable(entID)
	_doesExist(entID, "getting a entity objects collision state")
	return entities[entID][12]
end

function entityManager_module.getPosition(entID)
	_doesExist(entID, "getting a entity objects position")
	return entities[entID][1]
end

function entityManager_module.setCollidable(entID, val)
	_doesExist(entID, "setting a entity objects collision state")
	entities[entID][12] = val
end

function entityManager_module.setSprite(entID, spriteID)
	_doesExist(entID, "setting a entity objects sprite")
	entities[entID][3] = spriteID
end

function entityManager_module.setQuad(entID, quads)
	_doesExist(entID, "setting a entity objects quad table")
	if quads then
		entities[entID][4] = quads
		entities[entID][5] = 1
	else
		entities[entID][4] = nil
	end
end

function entityManager_module.quadSelect(entID, sel)
	_doesExist(entID, "selecting a quad")
	entities[entID][5] = sel
end

function entityManager_module.setTint(entID, r, g, b)
	_doesExist(entID, "setting a entity objects tint")
	entities[entID][6] = r or 0
	entities[entID][7] = g or 0
	entities[entID][8] = b or 0
end

function entityManager_module.setForce(entID, x, y)
	_doesExist(entID, "setting a entity objects force")
	entities[entID][9] = x or 0
	entities[entID][10]= y or 0
end

function entityManager_module.setAnchorOffset(entID, x, y)
	_doesExist(entID, "setting a entity objects anchor offset")
	entities[entID][14] = x or 0
	entities[entID][15] = y or 0
end

function entityManager_module.setPosition(entID, x, y)
	_doesExist(entID, "setting a entity objects position")
	entities[entID][1][1] = x
	entities[entID][1][2] = y
end

function entityManager_module.setRenderLayer(entID, layer)
	_doesExist(entID, "setting a entity objects render layer")
	entities[entID][19] = math.min(math.max(layer, 1), 10)
end

function entityManager_module.flip(entID, x, y)
	_doesExist(entID, "flipping an entity object")
	local newFlipX = x or false
	local newFlipY = y or false
	entities[entID][16] = newFlipX
	entities[entID][17] = newFlipY
	for i, v in pairs(entities[entID][11]) do
		entities[entID][18][i] = {boxFlip(v[1], v[2], v[3], v[4], newFlipX, newFlipY)}
	end
end

function entityManager_module.setBoundingBox(entID, boxNum, up, down, left, right)
	_doesExist(entID, "setting a entity objects bounding box")
	local newBB = entities[entID][11]
	local newFlipBB = entities[entID][18]
	if boxNum then
		local newBox = nil
		local newFlipBox = nil
		if up and down and left and right then
			newBox = {up, down, left, right}
			newFlipBox = {boxFlip(up, down, left, right, entities[entID][16], entities[entID][17])}
		end
		newBB[boxNum] = newBox
		newFlipBB[boxNum] = newFlipBox
	else
		newBB = {}
		newFlipBB = {}
	end
	entities[entID][11] = newBB
	entities[entID][18] = newFlipBB
end

function entityManager_module.anchor(ent1ID, ent2ID)
	if not entities[ent1ID] then
		u_error("anchoring two entity objects", "entity 1 does not exist!")
	end
	if not entities[ent2ID] then
		u_error("anchoring two entity objects", "entity 2 does not exist!")
	end
	if anchorTable[ent1ID] then
		u_error("anchoring two entity objects", "entity 1 is already anchored!")
	end
	if entities[ent2ID][13] then
		u_error("anchoring two entity objects", "entity 2 is already anchored!")
	end
	anchorTable[ent1ID] = ent2ID
	entities[ent2ID][13] = ent1ID
end

function entityManager_module.unanchor(entID)
	_doesExist(entID, "unanchoring a entity object")
	local anchorerID = nil
	local anchoreeID = nil
	if anchorTable[entID] then -- anchorer
		anchorerID = entID
		anchoreeID = anchorTable[entID]
	elseif entities[entID][13] then -- anchoree
		anchorerID = entities[entID][13]
		anchoreeID = entID
	else -- failsafe
		return
	end
	anchorTable[anchorerID] = nil
	entities[anchoreeID][13] = nil
end

function entityManager_module.entityCollision(ent1ID, ent2ID)
	if not entities[ent1ID] then
		u_error("checking collision between two entity objects", "entity 1 does not exist!")
	end
	if not entities[ent2ID] then
		u_error("checking collision between two entity objects", "entity 2 does not exist!")
	end
	local hasCollided = false
	puul(entities[ent1ID][1][1], entities[ent1ID][1][2], entities[ent1ID][18],
	entities[ent2ID][1][1], entities[ent2ID][1][2], entities[ent2ID][18], function(_)
		hasCollided = true
		return true
	end)
	return hasCollided
end

function entityManager_module.mapCollision(entID)
	_doesExist(entID, "testing map collision on a entity object")
	for i, v in pairs(entities[entID][18]) do
		if entities[entID][1][1] + v[3] < 0 then -- left
			return true
		elseif entities[entID][1][1] + v[4] > mapWidth then -- right
			return true
		end
		if entities[entID][1][2] + v[1] < 0 then -- up
			return true
		elseif entities[entID][1][2] + v[2] > mapHeight then -- down
			return true
		end
	end
	local isCollided = false
	iterateTiles(entities[entID][1], entities[entID][18], function()
		isCollided = true
		return true
	end, mapLoader_module.getForeMap())
	return isCollided
end

function entityManager_module.update()
	local foreMap = mapLoader_module.getForeMap()
	for entID, ent in pairs(entities) do
		-- movement
		if ent[13] then -- anchored
			-- there's nothing here because the anchorees position will be updated when the anchorer is being processed
		elseif ent[12] then -- normal
			ent[1][1] = ent[1][1] + ent[9] -- X
			scanPush(1, ent[1], ent[18], foreMap)
			ent[1][2] = ent[1][2] + ent[10] -- Y
			scanPush(2, ent[1], ent[18], foreMap)
			-- world bounds
			for i, v in pairs(ent[18]) do
				if ent[1][1] + v[3] < 0 then -- left
					ent[1][1] = ent[1][1] - (ent[1][1] + v[3])
				elseif ent[1][1] + v[4] > mapWidth then -- right
					ent[1][1] = ent[1][1] - (ent[1][1] + v[4] - mapWidth)
				end
				if ent[1][2] + v[1] < 0 then -- up
					ent[1][2] = ent[1][2] - (ent[1][2] + v[1])
				elseif ent[1][2] + v[2] > mapHeight then -- down
					ent[1][2] = ent[1][2] - (ent[1][2] + v[2] - mapHeight)
				end
			end
			if anchorTable[entID] then
				local anchoreeID = anchorTable[entID]
				local anchoreePos = entities[anchoreeID][1]
				anchoreePos[1] = ent[1][1] + entities[anchoreeID][14]
				anchoreePos[2] = ent[1][2] + entities[anchoreeID][15]
				entities[anchoreeID][1] = anchoreePos
			end
		else -- collision disabled
			ent[1][1] = ent[1][1] + ent[9] -- X
			ent[1][2] = ent[1][2] + ent[10] -- Y
		end
		ent[9] = 0
		ent[10]= 0
	end
end

local tintStore = {0, 0, 0}
local texSizeStore = {0, 0}
local viewportStore = {0, 0, 0, 0}
if multiCanvasLim == 1 then
	print("Using multi-pass entity rendering")
	function entityManager_module.render()
		love.graphics.setCanvas(stencilCanvas, renderCanvas)
		love.graphics.clear()
		love.graphics.setCanvas(renderCanvas)
		shaderManager_module.setShader("entityShader")
		for _, v in pairs(entities) do
			if v[3] ~= 0 then
				local pos = v[1]
				local spr = spritePool_module.get(v[3])
				local sprW, sprH = spr:getDimensions()
				-- set shader
				tintStore[1], tintStore[2], tintStore[3] = v[6], v[7], v[8]
				texSizeStore[1], texSizeStore[2] = sprW, sprH
				shaderManager_module.send("tint", tintStore) -- tint
				shaderManager_module.send("flipH", v[16]) -- flip
				shaderManager_module.send("flipV", v[17])
				shaderManager_module.send("texSize", texSizeStore)
				shaderManager_module.send("self", renderCanvas) -- layer
				shaderManager_module.send("stencil", stencilCanvas)
				shaderManager_module.send("layer", v[19] / 255)
				shaderManager_module.send("mode", false) -- mode
				-- draw
				love.graphics.push("all")
				love.graphics.setCanvas(stencilCanvas)
				love.graphics.setBlendMode("lighten", "premultiplied")
				if v[4] then
					local quad = v[4][v[5]]
					viewportStore[1], viewportStore[2], viewportStore[3], viewportStore[4] = quad:getViewport()
					shaderManager_module.send("viewPort", viewportStore)
					love.graphics.draw(spr, quad, pos[1], pos[2]) -- draw to stencil
					love.graphics.pop() -- pop and switch mode
					shaderManager_module.send("mode", true)
					love.graphics.draw(spr, quad, pos[1], pos[2]) -- draw to render
				else
					viewportStore[1], viewportStore[2], viewportStore[3], viewportStore[4] = 0, 0, sprW, sprH
					shaderManager_module.send("viewPort", viewportStore)
					love.graphics.draw(spr, pos[1], pos[2])
					love.graphics.pop()
					shaderManager_module.send("mode", true)
					love.graphics.draw(spr, pos[1], pos[2])
				end
			end
		end
		shaderManager_module.setShader()
		love.graphics.setCanvas()
		love.graphics.push()
		love.graphics.origin()
		love.graphics.draw(renderCanvas)
		love.graphics.pop()
	end
else
	print("Using single-pass entity rendering")
	function entityManager_module.render()
		love.graphics.setCanvas(stencilCanvas, renderCanvas)
		love.graphics.clear()
		shaderManager_module.setShader("entityShader")
		for _, v in pairs(entities) do
			if v[3] ~= 0 then
				local pos = v[1]
				local spr = spritePool_module.get(v[3])
				local sprW, sprH = spr:getDimensions()
				-- set shader
				tintStore[1], tintStore[2], tintStore[3] = v[6], v[7], v[8]
				texSizeStore[1], texSizeStore[2] = sprW, sprH
				shaderManager_module.send("tint", tintStore) -- tint
				shaderManager_module.send("flipH", v[16]) -- flip
				shaderManager_module.send("flipV", v[17])
				shaderManager_module.send("texSize", texSizeStore)
				shaderManager_module.send("self", renderCanvas) -- layer
				shaderManager_module.send("stencil", stencilCanvas)
				shaderManager_module.send("layer", v[19] / 255)
				-- draw
				if v[4] then
					local quad = v[4][v[5]]
					viewportStore[1], viewportStore[2], viewportStore[3], viewportStore[4] = quad:getViewport()
					shaderManager_module.send("viewPort", viewportStore)
					love.graphics.draw(spr, quad, pos[1], pos[2])
				else
					viewportStore[1], viewportStore[2], viewportStore[3], viewportStore[4] = 0, 0, sprW, sprH
					shaderManager_module.send("viewPort", viewportStore)
					love.graphics.draw(spr, pos[1], pos[2])
				end
			end
		end
		shaderManager_module.setShader()
		love.graphics.setCanvas()
		love.graphics.push()
		love.graphics.origin()
		love.graphics.draw(renderCanvas)
		-- love.graphics.draw(stencilCanvas)
		love.graphics.pop()
	end
end

return entityManager_module
