local itemManager_module = {}

local entityManager_module = require("modules/entity/entityManager")
local fileSet_module = require("modules/file/fileSet")
local spritePool_module = require("modules/graphics/spritePool")
local timerManager_module = require("modules/game/timerManager")

local u_error = require("modules/utils/error")
local u_createQuads = require("modules/utils/createQuads")
local u_loadModuleFiles = require("modules/utils/loadModuleFiles")

--[[
	item container:
	{entityID, item entry, is in play field, item type, item values, variables, pickup timer ID, despawn timer ID}
--]]

local ITEM_TYPES_FILEPATH = "modules/game/itemTypes"
local ITEM_PICK_UP_DELAY = 2.1
local ITEM_DESPAWN_TIME = 45;

local items = {}
local itemUseFunc = {}
local itemEquipFunc = {}
local itemNewFunc = {}
local itemEquipHoldFunc = {}

local function _doesExist(id, m)
	if not items[id] then
		u_error(m, "item object does not exist!")
	end
end

local function createItemEntity(itemSet, x, y)
	local entID = entityManager_module.new(x, y, itemSet[1])
	local entQuads = u_createQuads(16, 16, spritePool_module.getDimensions(itemSet[1]))
	entityManager_module.setQuad(entID, entQuads)
	for i, v in pairs(itemSet[2]) do
		entityManager_module.setBoundingBox(entID, i, v[1], v[2], v[3], v[4])
	end
	return entID
end

local function functionHandler(func, itemID, itemEntry, playerID, itemValues, itemVariables)
	local args = {func(itemEntry, playerID, itemValues, itemVariables)}
	local status = table.remove(args, 1)
	if status == 0 then -- destroy item
		itemManager_module.destroy(itemID)
	elseif status == 1 then -- return values to caller
		return args
	end
end

function itemManager_module.loadItemTypes()
	itemUseFunc = u_loadModuleFiles(ITEM_TYPES_FILEPATH, "lua", "item type", function(fileName, typeName)
		local functionTable = require(fileName)
		if functionTable["equip"] then
			itemEquipFunc[typeName] = functionTable["equip"]
		end
		if functionTable["new"] then
			itemNewFunc[typeName] = functionTable["new"]
		end
		if functionTable["equipHold"] then
			itemEquipHoldFunc[typeName] = functionTable["equipHold"]
		end
		if functionTable["use"] then
			return functionTable["use"]
		end
		u_error("loading an item type", "use function not found!")
	end)
end

function itemManager_module.new(itemEntry, x, y, inPlayField)
	local item = {}
	local newPlayField = inPlayField or true
	local entID = nil
	local itemSet = fileSet_module.get("item", itemEntry)
	if newPlayField then
		entID = createItemEntity(itemSet, x, y)
	end
	local itemValues = {}
	for i = 4, #itemSet do
		itemValues[#itemValues + 1] = itemSet[i]
	end
	item[1] = entID -- entity ID
	item[2] = itemEntry -- item entry
	item[3] = newPlayField -- is in play field
	item[4] = itemSet[3] -- item type
	item[5] = itemValues -- item values
	item[6] = {} -- variable
	item[7] = timerManager_module.new(ITEM_PICK_UP_DELAY, true) -- pickup timer ID
	item[8] = timerManager_module.new(ITEM_DESPAWN_TIME) -- despawn timer ID
	local newID = #items + 1
	items[newID] = item
	local func = itemNewFunc[item[4]]
	if func then
		return newID, functionHandler(func, #items, item[2], 0, item[5], item[6])
	end
	return newID
end

function itemManager_module.destroy(id)
	_doesExist(id, "destroying a item object")
	if items[id][3] then
		entityManager_module.destroy(items[id][1])
	end
	timerManager_module.destroy(items[id][7])
	timerManager_module.destroy(items[id][8])
	items[id] = nil
end

function itemManager_module.clear()
	for i, _ in pairs(items) do
		itemManager_module.destroy(i)
	end
end

function itemManager_module.getArray()
	return items
end

function itemManager_module.getEntry(id)
	return items[id]
end

function itemManager_module.getEntityID(id)
	_doesExist(id, "getting a item objects entity ID")
	return items[id][1]
end

function itemManager_module.doesExist(id)
	return items[id] ~= nil
end

function itemManager_module.isInPlayField(id)
	_doesExist(id, "checking if a item object is in the play field")
	return items[id][3]
end

function itemManager_module.removeFromPlayField(id)
	_doesExist(id, "removing a item object from the play field")
	if items[id][3] then
		entityManager_module.destroy(items[id][1])
		timerManager_module.reset(items[id][7])
		timerManager_module.reset(items[id][8])
		items[id][1] = nil
		items[id][3] = false
	end
end

function itemManager_module.insertIntoPlayField(id, x, y)
	_doesExist(id, "inserting a item object into the play field")
	if not items[id][3] then
		items[id][1] = createItemEntity(fileSet_module.get("item", items[id][2]), x, y)
		items[id][3] = true
		timerManager_module.resume(items[id][7])
		timerManager_module.resume(items[id][8])
	end
end

function itemManager_module.use(id, playerID)
	_doesExist(id, "calling an item objects use function")
	local func = itemUseFunc[items[id][4]]
	if func then
		return functionHandler(func, id, items[id][2], playerID or 0, items[id][5], items[id][6])
	else
		print("WARNING!\nFunction of item type "..items[id][4].." is not found")
	end
end

function itemManager_module.useEquip(id, playerID)
	_doesExist(id, "calling an item objects equip function")
	local func = itemEquipFunc[items[id][4]]
	if func then
		return functionHandler(func, id, items[id][2], playerID or 0, items[id][5], items[id][6])
	else
		print("WARNING!\nEquip function of item type "..items[id][4].." is not found")
	end
end

function itemManager_module.useEquipHold(id, playerID)
	_doesExist(id, "calling an item objects equip hold function")
	local func = itemEquipHoldFunc[items[id][4]]
	if func then
		return functionHandler(func, id, items[id][2], playerID or 0, items[id][5], items[id][6])
	else
		print("WARNING!\nEquip hold function of item type "..items[id][4].." is not found")
	end
end

function itemManager_module.setVariable(id, name, value)
	_doesExist(id, "setting an item objects variable")
	items[id][6][name] = value
end

function itemManager_module.getVariable(id, name)
	_doesExist(id, "getting an item objects variable")
	return items[id][6][name]
end

function itemManager_module.getPickupTimerID(id)
	_doesExist(id, "getting a timer object from a item object")
	return items[id][7]
end

function itemManager_module.update()
	for i, v in pairs(items) do
		if timerManager_module.getState(v[8]) == 2 then
			itemManager_module.destroy(i)
		end
	end
end

return itemManager_module