local playerManager_module = {}

local spritePool_module = require("modules/graphics/spritePool")
local entityManager_module = require("modules/entity/entityManager")
local keyboard_module = require("modules/input/keyboard")
local inventoryManager_module = require("modules/game/inventoryManager")
local itemManager_module = require("modules/game/itemManager")
local timerManager_module = require("modules/game/timerManager")
local sound_module = require("modules/sound/sound")

local u_error = require("modules/utils/error")
local u_decodeDirection = require("modules/utils/decodeDirection")
local u_createQuads = require("modules/utils/createQuads")

--[[
	player container:
	{entityID, {key inputs}, inventoryID, animation state, animation time, is in inventory mode, inventory index, health, max health,
	armour equip, weapon equip, direction, use cooldown}
	key inputs: up, down, left, right, inventory, attack, aim
	direction: left, nil, right, up left, up, up right, down left, down, down right
--]]

local PLAYER_BASE_SPEED = 1
local PLAYER_ANIMATION_SPEED = 0.246
local PLAYER_STARTING_HEALTH = 500

local players = {}

local playerSpriteSheetID = nil
local playerSpriteQuads = nil

local ready = false

local function _isPrepared(m, id)
	if not ready then
		u_error(m, "Player manager is not prepared!")
	end
	if id and not players[id] then
		u_error(m, "Player object does not exist!")
	end
end

local function getterHelper(functionName, valueID, errExt)
	playerManager_module["get"..functionName] = function(id)
		_isPrepared("getting a player objects "..errExt, id)
		return players[id][valueID]
	end
end

local function setterHelper(functionName, valueID, errExt)
	playerManager_module["set"..functionName] = function(id, val)
		_isPrepared("setting a player objects "..errExt, id)
		players[id][valueID] = val
	end
end

function playerManager_module.prepare()
	playerSpriteSheetID = spritePool_module.load("res/playerSprite.png")
	playerSpriteQuads = u_createQuads(16, 16, spritePool_module.getDimensions(playerSpriteSheetID))
	ready = true
end

function playerManager_module.finish()
	spritePool_module.unload(playerSpriteSheetID)
	playerSpriteSheetID = nil
	for i = 1, #playerSpriteQuads do
		playerSpriteQuads[i]:release()
	end
	playerSpriteQuads = nil
	ready = false
end

function playerManager_module.new(spawnPos, color, keyInput, disableInventory, maxHealth)
	_isPrepared("creating a player")
	local newColor = color or {0, 0, 0}
	local player = {}
	local entID = entityManager_module.new(spawnPos[1], spawnPos[2], playerSpriteSheetID)
	entityManager_module.setQuad(entID, playerSpriteQuads)
	entityManager_module.setTint(entID, newColor[1], newColor[2], newColor[3])
	entityManager_module.setBoundingBox(entID, 1, 1, 15, 4, 11)
	entityManager_module.setRenderLayer(entID, 2)
	local newInvID = nil
	if not disableInventory then
		newInvID = inventoryManager_module.new()
	end
	local newHealth = maxHealth or PLAYER_STARTING_HEALTH
	player[1] = entID -- entity ID
	player[2] = keyInput -- key inputs
	player[3] = newInvID -- inventory ID
	player[4] = 0 -- animation state
	player[5] = PLAYER_ANIMATION_SPEED -- animation time
	player[6] = false -- is in inventory mode
	player[7] = 1 -- inventory index
	player[8] = newHealth -- health
	player[9] = newHealth -- max health
	player[10]= 0 -- armour equip
	player[11]= 0 -- weapon equip
	player[12]= 3 -- direction
	player[13]= 0 -- use cooldown
	local newID = #players + 1
	players[newID] = player
	return newID
end

function playerManager_module.destroy(id)
	_isPrepared("destroying a player object", id)
	entityManager_module.destroy(players[id][1])
	if players[id][3] then
		inventoryManager_module.destroy(players[id][3])
	end
	players[id] = nil
end

function playerManager_module.clear()
	for i, _ in pairs(players) do
		playerManager_module.destroy(i)
	end
end

function playerManager_module.getArray()
	return players
end

function playerManager_module.getEntry(id)
	return players[id]
end

getterHelper("EntityID",       1, "entity ID")
getterHelper("InventoryID",    3, "inventory ID")
getterHelper("InventoryIndex", 7, "inventory index")
getterHelper("Health",         8, "health")
getterHelper("MaxHealth",      9, "maximum health")
getterHelper("ArmourEquip",   10, "armour equip value")
getterHelper("WeaponEquip",   11, "weapon equip value")
getterHelper("Direction",     12, "direction")

setterHelper("Health",    8, "health")
setterHelper("MaxHealth", 9, "maximum health")

function playerManager_module.addHealth(id, val)
	_isPrepared("adding to a player objects health", id)
	if val < 0 then -- deduct health
		local armourValue = 0
		if players[id][10] ~= 0 then
			local itemID = inventoryManager_module.getitemID(players[id][3], players[id][10])
			local args = itemManager_module.useEquip(itemID, id)
			armourValue = args[1]
		end
		local armourDeduct = math.floor(-val / (1 - (armourValue / -7.51) * math.pi) ^ 2)
		players[id][8] = math.max(players[id][8] - armourDeduct, 0)
	elseif val > 0 then -- add health
		players[id][8] = math.min(players[id][8] + val, players[id][9])
	end
end

function playerManager_module.getItemIDFromArmourEquip(id)
	_isPrepared("getting a item ID from a player objects armour equip pointer", id)
	if players[id][10] == 0 then
		return 0
	end
	return inventoryManager_module.getitemID(players[id][3], players[id][10])
end

function playerManager_module.getItemIDFromWeaponEquip(id)
	_isPrepared("getting a item ID from a player objects weapon equip pointer", id)
	if players[id][11] == 0 then
		return 0
	end
	return inventoryManager_module.getitemID(players[id][3], players[id][11])
end

local fistItem = {"fist", 15, "const", 7}
local directionLookUp = {5, 5, 5, 7, 6, 7, 7, 6, 7}
function playerManager_module.update(dt)
	for playerID, v in pairs(players) do
		-- inventory toggle
		if not keyboard_module.isPressed(v[2][7]) and keyboard_module.isPressed(v[2][5]) then
			v[6] = not v[6]
		end
		-- item collection
		if not inventoryManager_module.isFull(v[3]) then
			local itemArray = itemManager_module.getArray()
			for itemID, i in pairs(itemArray) do
				if itemManager_module.isInPlayField(itemID) and -- if the item is in the play field
				entityManager_module.entityCollision(v[1], i[1]) and -- and the player entity and the item entity is touching
				timerManager_module.getState(itemManager_module.getPickupTimerID(itemID)) ~= 0 then -- and the pick up timer is not running
					inventoryManager_module.insertItem(v[3], itemID)
					sound_module.soundEffectsHelper("pickup")
				end
			end
		end
		if not v[6] then -- normal mode?
			-- player movement and direction
			local ySpeed = 0
			local xSpeed = 0
			local newDirection = 2
			if keyboard_module.isDown(v[2][1]) then -- up
				ySpeed = ySpeed - PLAYER_BASE_SPEED
				newDirection = newDirection + 3
			elseif keyboard_module.isDown(v[2][2]) then -- down
				ySpeed = ySpeed + PLAYER_BASE_SPEED
				newDirection = newDirection + 6
			end
			if keyboard_module.isDown(v[2][3]) then -- left
				xSpeed = xSpeed - PLAYER_BASE_SPEED
				newDirection = newDirection - 1
			elseif keyboard_module.isDown(v[2][4]) then -- right
				xSpeed = xSpeed + PLAYER_BASE_SPEED
				newDirection = newDirection + 1
			end
			if newDirection ~= 2 then
				v[12] = newDirection
			end
			if not keyboard_module.isDown(v[2][7]) then
				entityManager_module.setForce(v[1], xSpeed, ySpeed)
				-- animation code
				if ySpeed ~= 0 or xSpeed ~= 0 then
					if v[5] >= PLAYER_ANIMATION_SPEED then
						v[4] = (v[4] + 1) % 2
						v[5] = v[5] - PLAYER_ANIMATION_SPEED
					end
					v[5] = v[5] + dt
				else
					v[4] = -1
					v[5] = PLAYER_ANIMATION_SPEED
				end
				entityManager_module.quadSelect(v[1], 2 + v[4])
				entityManager_module.flip(v[1], false, false)
			else
				-- aim code
				entityManager_module.quadSelect(v[1], directionLookUp[v[12]])
				entityManager_module.flip(v[1], u_decodeDirection(v[12], true, false, false, false, true))
			end
			-- item use code
			local keyCode = keyboard_module.getKey(v[2][6])
			if keyCode == 1 then -- pressed state
				if v[13] <= 0 then
					if v[11] ~= 0 then -- item
						local itemID = inventoryManager_module.getitemID(v[3], v[11])
						local args = itemManager_module.useEquip(itemID, playerID)
						if args and args[1] == 0 then -- add cooldown
							v[13] = tonumber(args[2])
						end
					else -- punch
						sound_module.soundEffectsHelper("meleeShoot", 7, 14)
						meleeManager_module.new(playerID, fistItem, v[12])
						v[13] = 27
					end
				end
			elseif keyCode == 2 then -- held state
				if v[13] <= 0 then
					if v[11] ~= 0 then -- item
						local itemID = inventoryManager_module.getitemID(v[3], v[11])
						local args = itemManager_module.useEquipHold(itemID, playerID)
						if args and args[1] == 0 then -- add cooldown
							v[13] = tonumber(args[2])
						end
					end
				end
			end
			-- cooldown code
			v[13] = math.max(v[13] - 1, 0)
		else -- inventory mode
			-- cursor code
			if keyboard_module.isPressed(v[2][3]) then -- left
				v[7] = v[7] - 1 + 8 * math.max(-v[7] + 2, 0)
				-- uncomment the line below when bitwise operators are available
				-- v[7] = (v[7] - 2 & 7) + 1
			elseif keyboard_module.isPressed(v[2][4]) then -- right
				v[7] = v[7] % 8 + 1
				-- uncomment the line below when bitwise operators are available
				-- v[7] = (v[7] & 7) + 1
			end
			-- use code
			if keyboard_module.isPressed(v[2][6]) then
				local itemID = inventoryManager_module.getitemID(v[3], v[7])
				if itemID ~= 0 then -- if there is no item
					local args = itemManager_module.use(itemID, playerID)
					if args and args[1] == 0 then -- equip
						v[9 + args[2]] = v[9 + args[2]] == v[7] and 0 or v[7]
					end
				else
					v[11] = 0
				end
			end
			-- drop code
			if keyboard_module.isPressed(v[2][7]) and inventoryManager_module.getitemID(v[3], v[7]) ~= 0 then
				local pos = entityManager_module.getPosition(v[1])
				inventoryManager_module.ejectItem(v[3], v[7], pos[1], pos[2])
				if v[7] == v[10] then
					v[10] = 0
				elseif v[7] == v[11] then
					v[11] = 0
				end
			end
			-- set quad
			entityManager_module.quadSelect(v[1], 4)
		end
	end
end

return playerManager_module