local sound_module = {}

local u_error = require("modules/utils/error")

--[[
	channel container:
	{audio source, destroy on end, state}
	
	channel states: 0: stopped 1: playing 2: paused
--]]

local audioFiles = {}
local channels = {}
local alias = {}
local invAlias = {}

local function _doesExist(id, m)
	if not channels[id] then
		u_error(m, "channel object does not exist!")
	end
end

function sound_module.load(filePath, isStatic, altName)
	local newID = #audioFiles + 1
	audioFiles[newID] = love.audio.newSource(filePath, isStatic and "static" or "stream")
	if altName then
		alias[altName] = newID
		invAlias[newID] = altName
	end
	return newID
end

function sound_module.replace(id, filePath, isStatic)
	if type(id) == "string" then
		local realID = alias[id]
		if realID then
			audioFiles[realID]:release()
			audioFiles[realID] = love.audio.newSource(filePath, isStatic and "static" or "stream")
		else
			sound_module.load(filePath, isStatic, id)
		end
	else
		if audioFiles[id] then
			audioFiles[id]:release()
		end
		audioFiles[id] = love.audio.newSource(filePath, isStatic and "static" or "stream")
	end
end

function sound_module.unload(id)
	if audioFiles[id] then
		audioFiles[id]:release()
		audioFiles[id] = nil
		if invAlias[id] then
			alias[invAlias[id]] = nil
			invAlias[id] = nil
		end
	end
end

function sound_module.unloadAll()
	for i, _ in pairs(audioFiles) do
		sound_module.unload(i)
	end
end

function sound_module.isLoaded(id)
	if type(id) == "string" then
		return audioFiles[alias[id]] and true or false
	end
	return audioFiles[id] and true or false
end

function sound_module.new(audioID, destroyOnEnd, playNow, loop, pitch)
	local newObj = {}
	local audioSource = nil
	if audioID then
		local realID = audioID
		if type(audioID) == "string" then
			realID = alias[audioID]
		end
		audioSource = audioFiles[realID]:clone()
		if loop then
			audioSource:setLooping(true)
		end
		if pitch then
			audioSource:setPitch(pitch)
		end
	end
	newObj[1] = audioSource -- audio source
	newObj[2] = destroyOnEnd and true or false -- destroy on end
	newObj[3] = 0 -- state
	if playNow and audioID then
		newObj[1]:play()
		newObj[3] = 1
	end
	local newID = #channels + 1
	channels[newID] = newObj
	return newID
end

function sound_module.destroy(id)
	_doesExist(id, "Destroying a channel object")
	if channels[id][1] then
		channels[id][1]:release()
	end
	channels[id] = nil
end

function sound_module.clear()
	for i, _ in pairs(channels) do
		sound_module.destroy(i)
	end
end

function sound_module.getState(id)
	_doesExist(id, "Getting the state of a channel object")
	return channels[id][3]
end

function sound_module.getLooping(id)
	_doesExist(id, "Getting the looping state of a channel object")
	if channels[id][1] then
		return channels[id][1]:isLooping()
	end
end

function sound_module.setLooping(id, loop)
	_doesExist(id, "Setting the looping state of a channel object")
	if channels[id][1] then
		channels[id][1]:setLooping(loop)
	end
end

function sound_module.setPitch(id, newPitch)
	_doesExist(id, "Setting the pitch of a channel object")
	if channels[id][1] then
		channels[id][1]:setPitch(loop)
	end
end

function sound_module.setSource(id, sourceID)
	_doesExist(id, "Setting the audio source of a channel object")
	if channels[id][1] then
		channels[id][1]:stop()
		channels[id][1]:release()
		channels[id][3] = 0
	end
	if sourceID then
		local realID = sourceID
		if type(sourceID) == "string" then
			realID = alias[sourceID]
		end
		channels[id][1] = audioFiles[realID]:clone()
	end
end

function sound_module.play(id)
	_doesExist(id, "Playing a channel object")
	if channels[id][1] then
		channels[id][1]:play()
		channels[id][3] = 1
	end
end

function sound_module.stop(id)
	_doesExist(id, "Stopping a channel object")
	if channels[id][1] then
		channels[id][1]:stop()
		channels[id][3] = 0
	end
end

function sound_module.pause(id)
	_doesExist(id, "Pausing a channel object")
	if channels[id][1] then
		channels[id][1]:pause()
		channels[id][3] = 2
	end
end

function sound_module.playAll()
	for i, _ in pairs(channels) do
		sound_module.play(i)
	end
end

function sound_module.stopAll()
	for i, _ in pairs(channels) do
		sound_module.stop(i)
	end
end

function sound_module.pauseAll()
	for i, _ in pairs(channels) do
		sound_module.pause(i)
	end
end

function sound_module.soundEffectsHelper(id, minPitch, maxPitch)
	sound_module.new(id, true, true, false, math.random(minPitch or 5, maxPitch or 20) / 10)
end

local whitelist = {
	-- wav and mp3
	["wav"]=true,["mp3"]=true,
	-- ogg vorbis
	["ogg"]=true,["oga"]=true,["ogv"]=true,
	-- tracker
	["699"]=true,["amf"]=true,["ams"]=true,["dbm"]=true,["dmf"]=true,
	["dsm"]=true,["far"]=true,["it"] =true,["j2b"]=true,["mdl"]=true,
	["med"]=true,["mod"]=true,["mt2"]=true,["mtm"]=true,["okt"]=true,
	["psm"]=true,["s3m"]=true,["stm"]=true,["ult"]=true,["umx"]=true,
	["xm"] =true,
	-- midi
	["abc"]=true,["mid"]=true,["pat"]=true,
}
function sound_module.folderPlay(sourceID, channelID, ...)
	_doesExist(channelID, "Playing from a folder")
	local folders = {...}
	local randomPick = 0
	local filePick = nil
	local selectedFDI = nil
	local selectedFI = 0
	local selectedFIF = nil
	while true do
		if #folders <= 0 then
			return
		end
		randomPick = math.random(1, #folders)
		filePick = folders[randomPick]
		if love.filesystem.getInfo(filePick, "directory") then
			selectedFDI = love.filesystem.getDirectoryItems(filePick)
			while true do
				if #selectedFDI <= 0 then
					break
				end
				selectedFI = math.random(1, #selectedFDI)
				selectedFIF = selectedFDI[selectedFI]
				local _, _, ext = string.find(selectedFIF, "%.(%w+)$")
				if whitelist[ext:lower()] then
					sound_module.replace(sourceID, filePick.."/"..selectedFIF)
					sound_module.setSource(channelID, sourceID)
					channels[channelID][1]:setLooping(true)
					channels[channelID][1]:setPitch(1)
					sound_module.play(channelID)
					return
				end
				table.remove(selectedFDI, selectedFI)
			end
		end
		table.remove(folders, randomPick)
	end
end

function sound_module.update()
	for i, v in pairs(channels) do
		if v[3] == 1 and not v[1]:isPlaying() then -- destroy on end
			if v[2] then
				sound_module.destroy(i)
			else
				v[3] = 0
			end
		end
	end
end

return sound_module